@@ -15,7 +15,7 @@
 #include <memory>
 #include <stdexcept>
 
-#include "xtl/xclosure.hpp"
+#include <xtl/xclosure.hpp>
 
 #include "xstorage.hpp"
 
@@ -50,9 +50,13 @@ namespace xt
             using self_type = xbuffer_storage<CP, A>;
             using allocator_type = A;
             using value_type = typename allocator_type::value_type;
-            using reference = typename allocator_type::reference;
+            using reference = std::conditional_t<std::is_const<std::remove_pointer_t<std::remove_reference_t<CP>>>::value,
+                                  typename allocator_type::const_reference,
+                                  typename allocator_type::reference>;
             using const_reference = typename allocator_type::const_reference;
-            using pointer = typename allocator_type::pointer;
+            using pointer = std::conditional_t<std::is_const<std::remove_pointer_t<std::remove_reference_t<CP>>>::value,
+                                  typename allocator_type::const_pointer,
+                                  typename allocator_type::pointer>;
             using const_pointer = typename allocator_type::const_pointer;
             using size_type = typename allocator_type::size_type;
             using difference_type = typename allocator_type::difference_type;
@@ -84,9 +88,13 @@ namespace xt
             using self_type = xbuffer_owner_storage<CP, A>;
             using allocator_type = A;
             using value_type = typename allocator_type::value_type;
-            using reference = typename allocator_type::reference;
+            using reference = std::conditional_t<std::is_const<std::remove_pointer_t<std::remove_reference_t<CP>>>::value,
+                                  typename allocator_type::const_reference,
+                                  typename allocator_type::reference>;
             using const_reference = typename allocator_type::const_reference;
-            using pointer = typename allocator_type::pointer;
+            using pointer = std::conditional_t<std::is_const<std::remove_pointer_t<std::remove_reference_t<CP>>>::value,
+                                  typename allocator_type::const_pointer,
+                                  typename allocator_type::pointer>;
             using const_pointer = typename allocator_type::const_pointer;
             using size_type = typename allocator_type::size_type;
             using difference_type = typename allocator_type::difference_type;
@@ -144,12 +152,14 @@ namespace xt
     public:
 
         using base_type = detail::buffer_storage_t<CP, A, O>;
+        using self_type = xbuffer_adaptor<CP, O, A>;
         using allocator_type = typename base_type::allocator_type;
         using value_type = typename base_type::value_type;
         using reference = typename base_type::reference;
         using const_reference = typename base_type::const_reference;
         using pointer = typename base_type::pointer;
         using const_pointer = typename base_type::const_pointer;
+        using temporary_type = uvector<value_type, allocator_type>;
 
         using size_type = typename base_type::size_type;
         using difference_type = typename base_type::difference_type;
@@ -164,6 +174,16 @@ namespace xt
         template <class P>
         xbuffer_adaptor(P&& data, size_type size, const allocator_type& alloc = allocator_type());
 
+        ~xbuffer_adaptor() = default;
+
+        xbuffer_adaptor(const self_type&) = default;
+        self_type& operator=(const self_type&) = default;
+
+        xbuffer_adaptor(self_type&&) = default;
+        xbuffer_adaptor& operator=(self_type&&) = default;
+
+        self_type& operator=(temporary_type&&);
+
         bool empty() const noexcept;
         using base_type::size;
         using base_type::resize;
@@ -225,6 +245,25 @@ namespace xt
     void swap(xbuffer_adaptor<CP, O, A>& lhs,
               xbuffer_adaptor<CP, O, A>& rhs) noexcept;
 
+    /************************************
+     * temporary_container metafunction *
+     ************************************/
+
+    template <class C>
+    struct temporary_container
+    {
+        using type = C;
+    };
+
+    template <class CP, class O, class A>
+    struct temporary_container<xbuffer_adaptor<CP, O, A>>
+    {
+        using type = typename xbuffer_adaptor<CP, O, A>::temporary_type;
+    };
+
+    template <class C>
+    using temporary_container_t = typename temporary_container<C>::type;
+
     /**********************************
      * xbuffer_storage implementation *
      **********************************/
@@ -401,6 +440,14 @@ namespace xt
     {
     }
 
+    template <class CP, class O, class A>
+    inline auto xbuffer_adaptor<CP, O, A>::operator=(temporary_type&& tmp) -> self_type&
+    {
+        base_type::resize(tmp.size());
+        std::copy(tmp.cbegin(), tmp.cend(), begin());
+        return *this;
+    }
+
     template <class CP, class O, class A>
     bool xbuffer_adaptor<CP, O, A>::empty() const noexcept
     {
