@@ -12,12 +12,12 @@
 #include <type_traits>
 #include <utility>
 
+#include <xtl/xcomplex.hpp>
+
 #include "xtensor/xbuilder.hpp"
 #include "xtensor/xexpression.hpp"
 #include "xtensor/xoffset_view.hpp"
 
-#include "xtl/xcomplex.hpp"
-
 namespace xt
 {
 
