@@ -15,6 +15,7 @@
 #include <utility>
 
 #include "xclosure.hpp"
+#include "xmeta_utils.hpp"
 #include "xtl_config.hpp"
 #include "xtype_traits.hpp"
 
@@ -78,10 +79,26 @@ namespace xtl
             using type = std::conditional_t < is_xoptional_impl<T>::value , T, xoptional<T >> ;
         };
 
+        template <class T>
+        struct identity
+        {
+            using type = T;
+        };
+
+        template <class T>
+        struct get_value_type
+        {
+            using type = typename T::value_type;
+        };
+
         template<class T1, class T2>
         struct common_optional_impl<T1, T2>
         {
-            using type = xoptional<std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>>;
+            using decay_t1 = std::decay_t<T1>;
+            using decay_t2 = std::decay_t<T2>;
+            using type1 = xtl::mpl::eval_if_t<std::is_fundamental<decay_t1>, identity<decay_t1>, get_value_type<decay_t1>>;
+            using type2 = xtl::mpl::eval_if_t<std::is_fundamental<decay_t2>, identity<decay_t2>, get_value_type<decay_t2>>;
+            using type = xoptional<std::common_type_t<type1, type2>>;
         };
 
         template <class T1, class T2, class B2>
@@ -119,13 +136,13 @@ namespace xtl
     using disable_xoptional = std::enable_if_t<!is_xoptional<E>::value, R>;
 
     template <class E1, class E2, class R = void>
-    using disable_both_xoptional = std::enable_if<!is_xoptional<E1>::value && !is_xoptional<E2>::value, R>;
+    using disable_both_xoptional = std::enable_if_t<!is_xoptional<E1>::value && !is_xoptional<E2>::value, R>;
 
     template <class E, class R = void>
     using enable_xoptional = std::enable_if_t<is_xoptional<E>::value, R>;
 
     template <class E1, class E2, class R = void>
-    using enable_both_xoptional = std::enable_if<is_xoptional<E1>::value && is_xoptional<E2>::value, R>;
+    using enable_both_xoptional = std::enable_if_t<is_xoptional<E1>::value && is_xoptional<E2>::value, R>;
 
     template <class... Args>
     struct common_optional : detail::common_optional_impl<Args...>
@@ -290,7 +307,6 @@ namespace xtl
         std::enable_if_t<
           conjunction<
             negation<std::is_same<xoptional<CT, CB>, std::decay_t<T>>>,
-            std::is_constructible<std::decay_t<CT>, T>,
             std::is_assignable<std::add_lvalue_reference_t<CT>, T>
           >::value,
          xoptional&>
@@ -304,7 +320,6 @@ namespace xtl
         template <class CTO, class CBO>
         std::enable_if_t<conjunction<
           negation<std::is_same<xoptional<CT, CB>, xoptional<CTO, CBO>>>,
-          std::is_constructible<std::decay_t<CT>, std::add_lvalue_reference_t<std::add_const_t<CTO>>>,
           std::is_assignable<std::add_lvalue_reference_t<CT>, CTO>,
           negation<detail::converts_from_xoptional<CT, CTO, CBO>>,
           negation<detail::assigns_from_xoptional<CT, CTO, CBO>>
@@ -320,7 +335,6 @@ namespace xtl
         template <class CTO, class CBO>
         std::enable_if_t<conjunction<
           negation<std::is_same<xoptional<CT, CB>, xoptional<CTO, CBO>>>,
-          std::is_constructible<std::decay_t<CT>, CTO>,
           std::is_assignable<std::add_lvalue_reference_t<CT>, CTO>,
           negation<detail::converts_from_xoptional<CT, CTO, CBO>>,
           negation<detail::assigns_from_xoptional<CT, CTO, CBO>>
@@ -342,6 +356,14 @@ namespace xtl
         xoptional& operator*=(const xoptional<CTO, CBO>&);
         template <class CTO, class CBO>
         xoptional& operator/=(const xoptional<CTO, CBO>&);
+        template <class CTO, class CBO>
+        xoptional& operator%=(const xoptional<CTO, CBO>&);
+        template <class CTO, class CBO>
+        xoptional& operator&=(const xoptional<CTO, CBO>&);
+        template <class CTO, class CBO>
+        xoptional& operator|=(const xoptional<CTO, CBO>&);
+        template <class CTO, class CBO>
+        xoptional& operator^=(const xoptional<CTO, CBO>&);
 
         template <class T>
         disable_xoptional<T, xoptional&> operator+=(const T&);
@@ -351,6 +373,14 @@ namespace xtl
         disable_xoptional<T, xoptional&> operator*=(const T&);
         template <class T>
         disable_xoptional<T, xoptional&> operator/=(const T&);
+        template <class T>
+        disable_xoptional<T, xoptional&> operator%=(const T&);
+        template <class T>
+        disable_xoptional<T, xoptional&> operator&=(const T&);
+        template <class T>
+        disable_xoptional<T, xoptional&> operator|=(const T&);
+        template <class T>
+        disable_xoptional<T, xoptional&> operator^=(const T&);
 
         // Access
         std::add_lvalue_reference_t<CT> value() & noexcept;
@@ -392,6 +422,58 @@ namespace xtl
         CB m_flag;
     };
 
+    // value
+
+    template <class T, class U = disable_xoptional<std::decay_t<T>>>
+    T&& value(T&& v)
+    {
+        return std::forward<T>(v);
+    }
+
+    template <class CT, class CB>
+    decltype(auto) value(xtl::xoptional<CT, CB>&& v)
+    {
+        return std::move(v).value();
+    }
+
+    template <class CT, class CB>
+    decltype(auto) value(xtl::xoptional<CT, CB>& v)
+    {
+        return v.value();
+    }
+
+    template <class CT, class CB>
+    decltype(auto) value(const xtl::xoptional<CT, CB>& v)
+    {
+        return v.value();
+    }
+
+    // has_value
+
+    template <class T, class U = disable_xoptional<std::decay_t<T>>>
+    bool has_value(T&&)
+    {
+        return true;
+    }
+
+    template <class CT, class CB>
+    decltype(auto) has_value(xtl::xoptional<CT, CB>&& v)
+    {
+        return std::move(v).has_value();
+    }
+
+    template <class CT, class CB>
+    decltype(auto) has_value(xtl::xoptional<CT, CB>& v)
+    {
+        return v.has_value();
+    }
+
+    template <class CT, class CB>
+    decltype(auto) has_value(const xtl::xoptional<CT, CB>& v)
+    {
+        return v.has_value();
+    }
+
     /***************************************
      * optional and missing implementation *
      ***************************************/
@@ -496,6 +578,54 @@ namespace xtl
         return *this;
     }
 
+    template <class CT, class CB>
+    template <class CTO, class CBO>
+    auto xoptional<CT, CB>::operator%=(const xoptional<CTO, CBO>& rhs) -> xoptional&
+    {
+        m_flag = m_flag && rhs.m_flag;
+        if (m_flag)
+        {
+            m_value %= rhs.m_value;
+        }
+        return *this;
+    }
+
+    template <class CT, class CB>
+    template <class CTO, class CBO>
+    auto xoptional<CT, CB>::operator&=(const xoptional<CTO, CBO>& rhs) -> xoptional&
+    {
+        m_flag = m_flag && rhs.m_flag;
+        if (m_flag)
+        {
+            m_value &= rhs.m_value;
+        }
+        return *this;
+    }
+
+    template <class CT, class CB>
+    template <class CTO, class CBO>
+    auto xoptional<CT, CB>::operator|=(const xoptional<CTO, CBO>& rhs) -> xoptional&
+    {
+        m_flag = m_flag && rhs.m_flag;
+        if (m_flag)
+        {
+            m_value |= rhs.m_value;
+        }
+        return *this;
+    }
+
+    template <class CT, class CB>
+    template <class CTO, class CBO>
+    auto xoptional<CT, CB>::operator^=(const xoptional<CTO, CBO>& rhs) -> xoptional&
+    {
+        m_flag = m_flag && rhs.m_flag;
+        if (m_flag)
+        {
+            m_value ^= rhs.m_value;
+        }
+        return *this;
+    }
+
     template <class CT, class CB>
     template <class T>
     auto xoptional<CT, CB>::operator+=(const T& rhs) -> disable_xoptional<T, xoptional&>
@@ -540,6 +670,50 @@ namespace xtl
         return *this;
     }
 
+    template <class CT, class CB>
+    template <class T>
+    auto xoptional<CT, CB>::operator%=(const T& rhs) -> disable_xoptional<T, xoptional&>
+    {
+        if (m_flag)
+        {
+            m_value %= rhs;
+        }
+        return *this;
+    }
+
+    template <class CT, class CB>
+    template <class T>
+    auto xoptional<CT, CB>::operator&=(const T& rhs) -> disable_xoptional<T, xoptional&>
+    {
+        if (m_flag)
+        {
+            m_value &= rhs;
+        }
+        return *this;
+    }
+
+    template <class CT, class CB>
+    template <class T>
+    auto xoptional<CT, CB>::operator|=(const T& rhs) -> disable_xoptional<T, xoptional&>
+    {
+        if (m_flag)
+        {
+            m_value |= rhs;
+        }
+        return *this;
+    }
+
+    template <class CT, class CB>
+    template <class T>
+    auto xoptional<CT, CB>::operator^=(const T& rhs) -> disable_xoptional<T, xoptional&>
+    {
+        if (m_flag)
+        {
+            m_value ^= rhs;
+        }
+        return *this;
+    }
+
     // Access
     template <class CT, class CB>
     auto xoptional<CT, CB>::value() & noexcept -> std::add_lvalue_reference_t<CT>
@@ -814,6 +988,110 @@ namespace xtl
         return e1.has_value() ? e1.value() / e2 : missing<value_type>();
     }
 
+    template <class T1, class B1, class T2, class B2>
+    inline auto operator%(const xoptional<T1, B1>& e1, const xoptional<T2, B2>& e2) noexcept
+        -> xoptional<std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>>
+    {
+        using value_type = std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>;
+        return e1.has_value() && e2.has_value() ? e1.value() % e2.value() : missing<value_type>();
+    }
+
+    template <class T1, class T2, class B2>
+    inline auto operator%(const T1& e1, const xoptional<T2, B2>& e2) noexcept
+        -> disable_xoptional<T1, common_optional_t<T1, T2>>
+    {
+        using value_type = std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>;
+        return e2.has_value() ? e1 % e2.value() : missing<value_type>();
+    }
+
+    template <class T1, class B1, class T2>
+    inline auto operator%(const xoptional<T1, B1>& e1, const T2& e2) noexcept
+        -> disable_xoptional<T2, common_optional_t<T1, T2>>
+    {
+        using value_type = std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>;
+        return e1.has_value() ? e1.value() % e2 : missing<value_type>();
+    }
+
+    template <class T, class B>
+    inline auto operator~(const xoptional<T, B>& e) noexcept
+        -> xoptional<std::decay_t<T>>
+    {
+        using value_type = std::decay_t<T>;
+        return e.has_value() ? ~e.value() : missing<value_type>();
+    }
+
+    template <class T1, class B1, class T2, class B2>
+    inline auto operator&(const xoptional<T1, B1>& e1, const xoptional<T2, B2>& e2) noexcept
+        -> xoptional<std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>>
+    {
+        using value_type = std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>;
+        return e1.has_value() && e2.has_value() ? e1.value() & e2.value() : missing<value_type>();
+    }
+
+    template <class T1, class T2, class B2>
+    inline auto operator&(const T1& e1, const xoptional<T2, B2>& e2) noexcept
+        -> disable_xoptional<T1, common_optional_t<T1, T2>>
+    {
+        using value_type = std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>;
+        return e2.has_value() ? e1 & e2.value() : missing<value_type>();
+    }
+
+    template <class T1, class B1, class T2>
+    inline auto operator&(const xoptional<T1, B1>& e1, const T2& e2) noexcept
+        -> disable_xoptional<T2, common_optional_t<T1, T2>>
+    {
+        using value_type = std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>;
+        return e1.has_value() ? e1.value() & e2 : missing<value_type>();
+    }
+
+    template <class T1, class B1, class T2, class B2>
+    inline auto operator|(const xoptional<T1, B1>& e1, const xoptional<T2, B2>& e2) noexcept
+        -> xoptional<std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>>
+    {
+        using value_type = std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>;
+        return e1.has_value() && e2.has_value() ? e1.value() | e2.value() : missing<value_type>();
+    }
+
+    template <class T1, class T2, class B2>
+    inline auto operator|(const T1& e1, const xoptional<T2, B2>& e2) noexcept
+        -> disable_xoptional<T1, common_optional_t<T1, T2>>
+    {
+        using value_type = std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>;
+        return e2.has_value() ? e1 | e2.value() : missing<value_type>();
+    }
+
+    template <class T1, class B1, class T2>
+    inline auto operator|(const xoptional<T1, B1>& e1, const T2& e2) noexcept
+        -> disable_xoptional<T2, common_optional_t<T1, T2>>
+    {
+        using value_type = std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>;
+        return e1.has_value() ? e1.value() | e2 : missing<value_type>();
+    }
+
+    template <class T1, class B1, class T2, class B2>
+    inline auto operator^(const xoptional<T1, B1>& e1, const xoptional<T2, B2>& e2) noexcept
+        -> xoptional<std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>>
+    {
+        using value_type = std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>;
+        return e1.has_value() && e2.has_value() ? e1.value() ^ e2.value() : missing<value_type>();
+    }
+
+    template <class T1, class T2, class B2>
+    inline auto operator^(const T1& e1, const xoptional<T2, B2>& e2) noexcept
+        -> disable_xoptional<T1, common_optional_t<T1, T2>>
+    {
+        using value_type = std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>;
+        return e2.has_value() ? e1 ^ e2.value() : missing<value_type>();
+    }
+
+    template <class T1, class B1, class T2>
+    inline auto operator^(const xoptional<T1, B1>& e1, const T2& e2) noexcept
+        -> disable_xoptional<T2, common_optional_t<T1, T2>>
+    {
+        using value_type = std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>;
+        return e1.has_value() ? e1.value() ^ e2 : missing<value_type>();
+    }
+
     template <class T1, class B1, class T2, class B2>
     inline auto operator||(const xoptional<T1, B1>& e1, const xoptional<T2, B2>& e2) noexcept
         -> xoptional<std::common_type_t<std::decay_t<T1>, std::decay_t<T2>>>
@@ -1119,6 +1397,12 @@ namespace xtl
     UNARY_OPTIONAL(erfc)
     UNARY_OPTIONAL(tgamma)
     UNARY_OPTIONAL(lgamma)
+    UNARY_OPTIONAL(ceil)
+    UNARY_OPTIONAL(floor)
+    UNARY_OPTIONAL(trunc)
+    UNARY_OPTIONAL(round)
+    UNARY_OPTIONAL(nearbyint)
+    UNARY_OPTIONAL(rint)
     UNARY_BOOL_OPTIONAL(isfinite)
     UNARY_BOOL_OPTIONAL(isinf)
     UNARY_BOOL_OPTIONAL(isnan)
